"""

    Key value store, with fast writing

"""

cdef extern from *:
    ctypedef char* const_char_ptr "const char*"
    ctypedef void** const_void_ptr_ptr "const void**"

cdef extern from "Python.h":
    int PyObject_AsReadBuffer(object obj, const_void_ptr_ptr buffer, Py_ssize_t *buffer_len)
    object PyString_FromStringAndSize(char *v, int len)
    
    ctypedef struct FILE
    FILE* PyFile_AsFile(object)
    void  fprintf(FILE* f, char* s, char* s)
    
    ctypedef unsigned long size_t
    FILE *fopen(char *path, char *mode)
    int fclose(FILE *fp)
    size_t fwrite(void *ptr, size_t size, size_t nmemb, FILE *stream)
    int fputc(int c, FILE *stream)
    long ftell(FILE *stream)
    
    int memcmp(void *s1, void *s2, size_t n)

ctypedef FILE *FILE_ptr
ctypedef char *char_ptr
    
import sys, os, mmap

try:
    import multiprocessing
except ImportError:
    import processing as multiprocessing
    
    #Hmm
    multiprocessing.Process.is_alive = multiprocessing.Process.isAlive




cdef inline int compare_strings(char *str1, long len1, char *str2, long len2):
    cdef long min_length
    if len1 < len2:
        min_length = len1
    else:
        min_length = len2
        
    cdef int result = memcmp(
        str1,
        str2,
        min_length
    )
        
    if not result:
        if len1 < len2:
            result = -1
        elif len1 > len2:
            result = 1
        
    return result


cdef inline void write_int64(FILE_ptr f, long value):
    fwrite( &value, 8, 1, f )

cdef inline void write_int(FILE_ptr f, long value):
    """ Write an unsigned integer, concisely """

    while value >= 128:
        fputc( (value&127)|128, f )
        value >>= 7
    fputc( value, f )

cdef inline long read_int(char **data):
    """ Read a concise unsigned integer """

    cdef long result = 0
    cdef unsigned char x
    cdef int shift = 0
    
    while True:
        x = data[0][0]
        data[0] += 1        
        result += (<long>(x&127)) << shift
        if (x&128) == 0: break
        shift += 7
    return result


cpdef remove_tree(path):
    """ Brittle and safe. """

    if not os.path.exists(path): return
    
    i = 0
    while True:
        filename = os.path.join(path,str(i))
        if not os.path.exists(filename): break
        os.unlink(filename)
        i += 1
    os.rmdir(path)


cpdef make_tree(path, sequence):
    cdef long i, j, ones, left_loc, right_loc
    cdef str key, value
    cdef int n_levels = 0
    cdef FILE_ptr files[64]    
    cdef long last[64]

    remove_tree(path)
    os.mkdir(path)
    
    i = 0
    for key, value in sequence:
        # The run of least significant ones determines the depth in the balanced tree
        ones = 0
        j = i
        while i & (1<<ones):
            ones += 1
        
        while n_levels <= ones:
            filename = os.path.join(path, str(n_levels))
            files[n_levels] = fopen(filename, 'wb')
            last[n_levels] = 0
            n_levels += 1            

        # last[ones] = levels[ones].tell()
        
        if ones:
            left_loc = last[ones-1]
            for j in xrange(ones):            
                last[j] = ftell(files[j])
            right_loc = last[ones-1]
            
            write_int(files[ones], left_loc)
            
            # Right should not be far from left,
            # write as a relative position to safe space
            write_int(files[ones], right_loc - left_loc)
            
        write_int(files[ones], len(key))        
        fwrite( <char*>key, len(key), 1, files[ones] )

        write_int(files[ones], len(value))
        fwrite( <char*>value, len(value), 1, files[ones] )
        
        i += 1

    if n_levels: 
        # The last little bits of the sequence might not be a part of the binary tree yet
        # so need to store their location
        # See also: Cursor.right()
    
        write_int64(files[0], 0)
        for i in xrange(0,n_levels-1):
            write_int64(files[i+1], last[i])
            
    for i in xrange(n_levels): 
        fclose(files[i])


cdef class Cursor

cdef class Tree:
    cdef public object path, levels
    cdef Py_ssize_t sizes[64]
    cdef char_ptr maps[64]

    def __init__(self, path):
        self.path = path
        self.levels = [ ]
        while True:
             filename = os.path.join(path, str(len(self.levels)))
             if not os.path.exists(filename): break
             f = open(filename,'rb')
             self.levels.append(mmap.mmap(f.fileno(), 0, prot=mmap.PROT_READ))
             f.close()
      
        for i in xrange(len(self.levels)):
            #self.sizes[i] = len(self.levels[i])
            PyObject_AsReadBuffer(self.levels[i], <const_void_ptr_ptr>&(self.maps[i]),
                <Py_ssize_t*>&(self.sizes[i]))

    cpdef close(self):
        for level in self.levels:
            level.close()

    cpdef size(self):
        return sum([ len(item) for item in self.levels ])

    cpdef Cursor cursor(self):
        return Cursor(self, len(self.levels)-1, 0, None)
        

cdef class Cursor:
    cdef public Tree tree
    cdef public Cursor successor
    cdef public long level

    cdef public long left_loc, right_loc, key_length
    cdef char *data_ptr

    def __cinit__(Cursor self, Tree tree, long level, long position, Cursor successor):
        self.tree = tree
        self.level = level
        #self.position = position
        self.successor = successor
        
        self.data_ptr = tree.maps[level] + position
        if level:
            self.left_loc = read_int(&self.data_ptr)
            self.right_loc = self.left_loc + read_int(&self.data_ptr)
        self.key_length = read_int(&self.data_ptr)

    cpdef int compare(Cursor self, Cursor other):
        return compare_strings(self.data_ptr,self.key_length,other.data_ptr,other.key_length)

    cpdef int compare_to_string(Cursor self, char *string, long length):
        return compare_strings(self.data_ptr,self.key_length,string,length)
        
    cpdef str key(Cursor self):
        return PyString_FromStringAndSize(self.data_ptr, self.key_length)
        
    cpdef str value(Cursor self):
        cdef char *value_ptr = self.data_ptr + self.key_length
        cdef long value_length = read_int(&value_ptr)
    
        return PyString_FromStringAndSize(value_ptr, value_length)

    cpdef Cursor left(Cursor self):
        if self.level == 0: return None
        
        return Cursor(self.tree, self.level-1, self.left_loc, self)

    cpdef Cursor right(Cursor self):
        if self.level == 0: return None

        cdef long new_level = self.level-1
        cdef long new_position = self.right_loc
        
        while new_position == self.tree.sizes[new_level] - 8:
            if new_level == 0: return None
        
            new_position = (<long*>(self.tree.maps[new_level]+new_position))[0]            
            new_level -= 1
        
        return Cursor(self.tree, new_level, new_position, 
                      self.successor)
    
    cpdef Cursor next(Cursor self):
        cdef Cursor right = self.right()
        if right is None: return self.successor
        return left_most(right)


cpdef Cursor left_most(Cursor cursor):
    if not cursor: return None
    while cursor.level > 0:
        cursor = cursor.left()
    return cursor


cdef class Cursor_merge(object):
    cdef Cursor cursor1, cursor2

    def __cinit__(self, cursor1, cursor2):
        self.cursor1 = cursor1
        self.cursor2 = cursor2

    def __iter__(self):
        return self

    def __next__(self):        
        if self.cursor1 is None:
            if self.cursor2 is None:
                raise StopIteration()
                
            result = (self.cursor2.key(), self.cursor2.value())
            self.cursor2 = self.cursor2.next()
            return result

        if self.cursor2 is None:
            result = (self.cursor1.key(), self.cursor1.value())
            self.cursor1 = self.cursor1.next()
            return result
        
        cdef int comparison = self.cursor1.compare(self.cursor2)
        
        if comparison < 0:
            result = (self.cursor1.key(), self.cursor1.value())
            self.cursor1 = self.cursor1.next()
            return result

        elif comparison > 0:
            result = (self.cursor2.key(), self.cursor2.value())
            self.cursor2 = self.cursor2.next()
            return result

        else:
            result = (self.cursor2.key(), self.cursor2.value())
            self.cursor1 = self.cursor1.next()
            self.cursor2 = self.cursor2.next()
            return result


cdef class Store:
    cdef public object path, trees, cursors, next_tree_id, memory, memory_size
    cdef public object lock, merger_processes, busy_paths
    
    def __init__(self, path, memory_size=1000000):
        self.path = path
        if not os.path.exists(path):
            os.mkdir(path)

        self.memory_size = memory_size
        self.memory = { }
        
        self.lock = multiprocessing.Lock()
        self.merger_processes = [ ]
        self.busy_paths = set()
        
        self.trees = [ ]
        self.remap()

        
    cpdef remap(self):
        self.lock.acquire()
        
        for tree in self.trees:
            tree.close()
    
        list_numbers = [ int(filename) 
                         for filename in os.listdir(self.path)
                         if filename.isdigit() ]
        list_numbers.sort()
        
        if not list_numbers:
            self.next_tree_id = 0
        else:
            self.next_tree_id = list_numbers[-1] + 1
    
        self.trees = [ ]
        for i in list_numbers:
            filename = os.path.join(self.path, str(i))            
            if not os.path.exists(filename): break
            self.trees.append( Tree(filename) )
                
        self.lock.release()
        
        self.cursors = [ tree.cursor() for tree in self.trees[::-1] ] #Note reverse order!
        
    
    cpdef clear(self):
        self.finish_merging()
    
        for tree in self.trees:
            remove_tree(tree.path)
        self.memory = { }
                
        self.remap()

    
    cpdef finish_merging(self, wait=True):
        still_waiting = [ ]
        any = False
        for item in self.merger_processes:
            if not wait and item[0].is_alive():
                still_waiting.append(item)
                continue
            
            #if item[0].is_alive(): print 'wait!'
            
            item[0].join()
            self.busy_paths.remove(item[1])
            self.busy_paths.remove(item[2])
            #print 'done', item[1], item[2]
            any = True
        
        self.merger_processes = still_waiting
        
        if any:
            self.remap()
            
        return any
        
    
    cpdef start_merging(self, all=False):
        self.finish_merging(False)
        
        pos = len(self.trees)-2
        while pos >= 0:
            if self.trees[pos].path not in self.busy_paths and \
               self.trees[pos+1].path not in self.busy_paths and \
               (all or self.trees[pos].size() <= self.trees[pos+1].size() * 2):
                path1 = self.trees[pos].path
                path2 = self.trees[pos+1].path
                self.busy_paths.add(path1)
                self.busy_paths.add(path2)
                #print 'merge', path1, path2
                
                process = multiprocessing.Process(target=do_mergedown,args=(self.path,path1,path2,self.lock))
                process.start()
                self.merger_processes.append((process,path1,path2))
                
                pos -= 2
            else:
                pos -= 1


    def optimize(self):
        self.flush()
    
        #Not optimal? Need a way to wait for any one process to finish.
        while len(self.trees) > 1:
            self.start_merging(True)
            self.finish_merging(True)        

    
    cpdef flush(self):
        if not self.memory: return
        
        tempname = os.path.join(self.path, 'newtree')
        filename = os.path.join(self.path, str(self.next_tree_id))

        # This seems to be about the fastest way to do this, surprisingly
        items = [ (item, self.memory[item]) for item in sorted(self.memory.iterkeys()) ]
        self.memory = {}
        
        make_tree(tempname, items)        
        os.rename(tempname, filename)
        self.remap()        
        self.start_merging()
        

    cpdef close(self):
        self.flush()
        self.finish_merging()        
        for tree in self.trees:
            tree.close()


    def __setitem__(self, key, value):
        self.memory[key] = value
        if len(self.memory) >= self.memory_size: 
            self.flush()       

    def __getitem__(self, key):
        cdef Cursor cursor

        if self.merger_processes:
            if self.finish_merging(False):
                self.start_merging()
        
        if key in self.memory:
            return self.memory[key]
            
        cdef char *key_ptr = key
        cdef long key_len = len(key)
        cdef int result
        
        for cursor in self.cursors:
            while cursor:
                result = cursor.compare_to_string(key_ptr, key_len)
                if result == 0: return cursor.value()
                if result > 0:
                    cursor = cursor.left()
                else:
                    cursor = cursor.right()
        raise KeyError(key)

def do_mergedown(path, path1, path2, lock):
    new_filename = path1 + '-new'
    
    lock.acquire()
    tree1 = Tree(path1)
    tree2 = Tree(path2)
    lock.release()
    
    make_tree(new_filename, 
        Cursor_merge(
            left_most(tree1.cursor()),
            left_most(tree2.cursor())))
    
    new_new_filename = path1
    
    lock.acquire()
    remove_tree(path1)
    remove_tree(path2)
    os.rename(new_filename, new_new_filename)
    lock.release()






